#!/usr/local/bin/perl5 -w 
# $Id: draw_chromosome_gc.pl,v 1.14 2003/04/01 08:00:23 hpaces Exp $

use GD;

###############################################################################
# init
###############################################################################

$beg = 0;               # for drawing only part of the sequence - not ready yet
$ch_name = "";
$scale = 10;
$window = 100000;
$gap_size = $window/10;
$nogap = 0;
$gap_number = 0;
$labels = 1;            # draw labels
$verbose = 1;           # verbose by default
$isochores = 1;         # gc5 default
$shift = 0;		  
			  
$verbose = $ENV{VERBOSE}+1 if defined($ENV{VERBOSE});
$verbose = $ENV{DEBUG}+1 if defined($ENV{DEBUG});
			  
$message = "Program for drawing large scale gc levels.\n\n";
$message .= "usage     : draw_chromosome_gc.pl [parameters] seq.fa png\n\n";
$message .= "-scale=#  : pixels / 1MB (default $scale)\n";
$message .= "-window=# : size of sliding window (default $window)\n";
$message .= "-gap=#    : minimum size of the gap to be drawn as a grey box\n";
$message .= "            [N in seq] (default $gap_size)\n";
$message .= "            -gap=0 or -nogap: no shading\n";
$message .= "-ch=#     : name of the chromosome for the picture label (max 2 characters)\n";
$message .= "-silent   : be really quiet\n";
$message .= "-nolabels : don't draw numbers on x and y axes\n";
$message .= "-gc5/-gc2.5/-isochores\n";
$message .= "   -gc5       : use 5% gc difference for color coding (default)\n";
$message .= "   -gc2.5     : use 2.5% gc difference for color coding\n";
$message .= "   -isochores : use isochores (37-42-46-52) for color coding\n";
$message .= "-shift=#  : shift gc range and colorcoding about #%\n";
$message .= "seq       : sequence (in fasta format)\n";
$message .= "png       : output png name\n\n";
$message .= "(export VERBOSE=# or DEBUG=# for more verbose output)\n";
			  
# check parameters	  
foreach (@ARGV) {	  
    if (/(\b|-)-h(\b|[elp])/) {
	die $message;	  
    } elsif (/^-scale=\d+$/) {
	($scale) = /-scale=(\d+)/ 
	    or die "$message\nError: can't resolve parameter $_\n"; 
    } elsif (/^-window=\d+$/) {
	($window) = /-window=(\d+)/ 
	    or die "$message\nError: can't resolve parameter $_\n"; 
    } elsif (/^-gap=\d+$/) {
	($gap_size) = /-gap=(\d+)/ 
	    or die "$message\nError: can't resolve parameter $_\n"; 
	$nogap = 1 unless $gap_size;
    } elsif (/^-nogap$/) {
	$nogap = 1;	  
    } elsif (/^-ch=.{1,2}$/) {
	($ch_name) = /-ch=(\S+)/
	    or die "$message\nError: can't resolve parameter $_\n"; 
    } elsif (/^-nolabels$/) {
	$labels = 0;	  
    } elsif (/^-gc5$/) {  
	$isochores = 1;   
    } elsif (/^-gc2\.5$/) {
	$isochores = 2;   
    } elsif (/^-isochores$/) {
	$isochores = 3;   
    } elsif (/^-shift=.+$/) {
	($shift) = /-shift=([\+|\-]{0,1}\d+\.{0,1}\d*)/
	    or die "$message\nError: can't resolve parameter $_\n"; 
    } elsif (/^-silent$/) {
	$verbose = 0; 	  
    } else {		  
	push @args, $_;	  
    };			  
};			  
$seq_file_name = shift @args or die $message;
$png_file_name = shift @args or die $message;
			  
###############################################################################
# parameters		      
# change parameters to fit your needs here 
###############################################################################
			      
# border (white space on sides in pixels)
$border = 35;		      
# space for x axis (in pixels)
$top = 25;		      
# png y size		      
$y = 250; # 250 default	      
# read sequence
$seq_len = &load_fasta_speed;
# calculating png x size 
$x = 2*$border + $seq_len/1000000*$scale;
print "Sequence is $seq_len nt long; setting xsize to $x pixels\n" 
    if $verbose;  	  
# $window/$step is one point; adjusting for different scales
$step = $scale*$window/200000; 
#$step = 10;
#$step *= 2 if $window/$scale<=50000;
#$step *= 2 if $window/$scale<=10000;
#$step *= 2 if $window/$scale<=5000;
#$step *= 2 if $window/$scale<=1000;
#$step *= 2 if $window/$scale<=500;
#$step /= 2 if $window/$scale>=100000;
#$step /= 2 if $window/$scale>=500000;
#$step /= 2 if $window/$scale>=1000000;
#$step /= 2 if $window/$scale>=5000000;
#$step /= 2 if $window/$scale>=10000000;
print "Debug: scale[$scale] window[$window] step[$step]\n" if $verbose>1;
			      
###############################################################################
# you should not change anything after this unless you really know, what you 
# are doing		      
# end of parameters	      
###############################################################################
		  	      
###############################################################################
# program	  	      
###############################################################################
		  	      
# calculate	  	  
&calc_gc;	  	  
($gap_number) = &find_gaps unless $nogap;
# initialize image, colors and font
&init_picture;	  	  
($isochores,$y_gc_proc_min,$y_gc_proc_max) = &init_colors($isochores);
# y convertor	  	  
$py = ($y-$top)/($y_gc_proc_max-$y_gc_proc_min);
# draw graph	  	  
print "Drawing ...\n" if $verbose;
&draw_color_graph;	  
&draw_graph;	  	  
# shade gaps not ready yet
&shade_gaps if $gap_number and not $nogap;
# draw isochores  	  
&draw_isochores;  	  
# axes		  	  
&draw_x_axis;	  	  
&draw_y_axis if $labels;  
&draw_ch_name if $ch_name;
		  	  
###############################################################################
# finish 	  	  
###############################################################################
		  	  
# Convert the image to png and print it to file
open (PNG, ">".$png_file_name) 
    or die "Error: can't open $png_file_name for writing. Check permitions\n";
print {PNG} $graph->png   
    or die "Error: can't write to $png_file_name. Check disk space\n";
close (PNG) 	  	  
    or die "Error: can't close $png_file_name. Something is wrong? Sorry.\n";
		  	  
print "Done.\n" if $verbose;
		  	  
###############################################################################
# end main	  	  
###############################################################################
		  	  
###############################################################################
# functions	  	  
###############################################################################
sub load_fasta_speed {	  
    my ($i,$s_info,@seq,$lines);
    		  	  
    print "Reading sequence ...\n" if $verbose;
# load as fasta file	  
    open SEQ, $seq_file_name    or die "Error: can't open $seq_file_name\n";
# first line is description
    $s_info = <SEQ>;	  
    die "Error: $seq_file_name is not in fasta format.\n" 
	unless $s_info =~ /^>/;
		  
    $i = 0;	  
    while ($seq[$i] = <SEQ>) {
# delete 'white' characters
	$seq[$i] =~ s/\s//g;
	$i++;	  
    };		  
    close SEQ or die "Error: can't close $seq_file_name.\n";
    pop @seq; # last field in array is undefined
    $seq = join "",@seq;
    $seq = uc $seq;
    $lines = @seq;
    print "... got ",length($seq)," nt in $lines lines\n" if $verbose>1;
    return length($seq);
};

##############################################################################
sub load_fasta_mem {
    my ($i,$s_info,@seq);
    
    print "Reading sequence ...\n" if $verbose;
# load as fasta file
    open SEQ, $seq_file_name    or die "Error: can't open $seq_file_name\n";
# first line is description
    $s_info = <SEQ>;
    die "Error: not $seq_file_name is not in fasta format.\n" 
	unless $s_info =~ /^>/;

    $i = 0;
    while ($seq[$i] = <SEQ>) {
# delete 'white' characters
	$seq[$i] =~ s/\s//g;
	$i++;
    };
    close SEQ or die "Error: can't close $seq_file_name.\n";

    $seq = join "",@seq;
    $seq = uc $seq;
    $lines = @seq;
    print "... got ",length($seq)," nt in $lines lines\n" if $verbose>1;
    return length($seq);
};

##############################################################################
# calc_gc
# calculating % GC in window (new code - don't count N's)
sub calc_gc {
    my($i,$ci,$gc,$at);
    
    print "Calculating ...\n" if $verbose;
    $ci=0;
    for ($i=0; $i<$seq_len-$window; $i+=int($window/$step)) {
	$gc = substr($seq,$i,$window) =~ tr/G|C//; 
	$at = substr($seq,$i,$window) =~ tr/A|T//; 
	if ($gc and $at) {
	    $data[$ci]=100*$gc/($at+$gc);
	} else {
	    $data[$ci]=0;
	};
	$ci++;
    };
};

###############################################################################
# find gaps
# find stretches on N's in row longer than $gap_size
sub find_gaps {
    my $i = -1;

    print "Finding gaps ...\n" if $verbose;
    while ($seq =~ /(N{$gap_size,})/g) {
	$i++;
	$gap_end[$i] = pos $seq;
	$gap_beg[$i] = $gap_end[$i] - length $&;
    };
    print "... found ",$i+1," gaps\n" if $verbose>1;
    return $i;
};

###############################################################################
# graphical functions
###############################################################################
sub init_picture {

    print "Initialize picture ...\n" if $verbose;
# create a new image
    $graph = new GD::Image($x,$y);

# allocate colors
    $white = $graph->colorAllocate(255,255,255);
    $black = $graph->colorAllocate(0,0,0);  
#  $red = $graph->colorAllocate(255,0,0);
    $grey = $graph->colorAllocate(230,230,230);
    
# make the background transparent and interlaced
#  $graph->transparent($white); $graph->interlaced('true');

# set font
    $font = gdMediumBoldFont;
    $w=$font->width; $h=$font->height;
};

###############################################################################
# init colors and gc levels
sub init_colors {
    my $isochores = shift;
    my ($y_gc_proc_min, $y_gc_proc_max);

# gc5    
    if ($isochores == 1) {

	$isochore_proc[0] = 37.5; 
	$isochore_proc[1] = 42.5; 
	$isochore_proc[2] = 47.5; 
	$isochore_proc[3] = 52.5;

 	$isochore_name[0] = "";
	$isochore_name[1] = "";
	$isochore_name[2] = "";
	$isochore_name[3] = "";
	$isochore_name[4] = "";

	$isochore_color[0] = $graph->colorAllocate(0,100,255);
	$isochore_color[1] = $graph->colorAllocate(0,200,255);
	$isochore_color[2] = $graph->colorAllocate(255,255,0);
	$isochore_color[3] = $graph->colorAllocate(255,130,0);
	$isochore_color[4] = $graph->colorAllocate(255,0,0);
	
	$isochores = 4;

# gc level to show (30-65 for isochores in human)
	$y_gc_proc_min = 30;
	$y_gc_proc_max = 65;

# gc2.5
    } elsif ($isochores == 2) {

	$isochore_proc[0] = 32.5; 
	$isochore_proc[1] = 35; 
	$isochore_proc[2] = 37.5; 
	$isochore_proc[3] = 40;
	$isochore_proc[4] = 42.5;
	$isochore_proc[5] = 45;
	$isochore_proc[6] = 47.5;
	$isochore_proc[7] = 50;
	$isochore_proc[8] = 52.5;
	$isochore_proc[9] = 55;

 	$isochore_name[0] = "";
	$isochore_name[1] = "";
	$isochore_name[2] = "";
	$isochore_name[3] = "";
	$isochore_name[4] = "";
	$isochore_name[5] = "";
	$isochore_name[6] = "";
	$isochore_name[7] = "";
	$isochore_name[8] = "";
	$isochore_name[9] = "";
	$isochore_name[10] = "";

	$isochore_color[0] = $graph->colorAllocate(0,0,255);
	$isochore_color[1] = $graph->colorAllocate(0,100,255);
	$isochore_color[2] = $graph->colorAllocate(0,200,255);
	$isochore_color[3] = $graph->colorAllocate(0,255,255);
	$isochore_color[4] = $graph->colorAllocate(100,255,0);
	$isochore_color[5] = $graph->colorAllocate(200,255,0);
	$isochore_color[6] = $graph->colorAllocate(255,255,0);
	$isochore_color[7] = $graph->colorAllocate(255,200,0);
	$isochore_color[8] = $graph->colorAllocate(255,150,0);
	$isochore_color[9] = $graph->colorAllocate(255,100,0);
	$isochore_color[10] = $graph->colorAllocate(255,0,0);
	
	$isochores = 10;
# gc level to show (30-65 for isochores in human)
	$y_gc_proc_min = 30;
	$y_gc_proc_max = 65;

# isochores
    } elsif ($isochores == 3) {

# $L1 = 37; $L2 = 41; $H1 = 46; $H2 = 52;
	$isochore_proc[0] = 37; 
	$isochore_proc[1] = 41; 
	$isochore_proc[2] = 46; 
	$isochore_proc[3] = 52;

 	$isochore_name[0] = "L1";
	$isochore_name[1] = "L2";
	$isochore_name[2] = "H1";
	$isochore_name[3] = "H2";
	$isochore_name[4] = "H3";

	$isochore_color[0] = $graph->colorAllocate(0,100,255);
	$isochore_color[1] = $graph->colorAllocate(0,200,255);
	$isochore_color[2] = $graph->colorAllocate(255,255,0);
	$isochore_color[3] = $graph->colorAllocate(255,130,0);
	$isochore_color[4] = $graph->colorAllocate(255,0,0);
	
	$isochores = 4;

# gc level to show (30-65 for isochores in human)
	$y_gc_proc_min = 30;
	$y_gc_proc_max = 65;
    };

# shift
    if ($shift) {
	for ($i=0; $i<$isochores; $i++) {
	    $isochore_proc[$i] += $shift;
	};
	$y_gc_proc_min += $shift;
	$y_gc_proc_max += $shift;
    };

    return $isochores,$y_gc_proc_min,$y_gc_proc_max;
};
	
###############################################################################
# x axis
sub draw_x_axis {
    my ($i,$position,$number,$shift);

# ticks and labels
    my $tick = 1000000;
    my $bigtick = 5000000;
    my $label = 10000000;

# adjust ticks for smaller ranges
    if ($seq_len < 1200000) {
	$tick = 500000;
	$bigtick = 1000000;
	$label = 500000;
    };
    if ($seq_len < 600000) {
	$tick = 100000;
	$bigtick = 500000;
	$label = 100000;
    };

# adjust ticks for whole chromosomes
    if ($seq_len > 15000000) {
	$tick = 5000000;
	$bigtick = 10000000;
	$label = 10000000;
    };

# main line
    $graph->line($border,20,$x-$border-1,20,$black);
# small crosses each $tick 
    for ($i=0; $i<=$seq_len; $i=$i+$tick) {
	$position=(($i-$beg)*($x-2*$border)/$seq_len)+$border;
	$graph->line($position,20,$position,15,$black);
    };
# big crosses each $tick_label
    for ($i=0; $i<=$beg+$seq_len; $i=$i+$bigtick) {
	$position=(($i-$beg)*($x-2*$border)/$seq_len)+$border;
	$graph->line($position,20,$position,13,$black);
    };
    if ($labels) {
# labels each $label
	for ($i=0; $i<=$beg+$seq_len; $i=$i+$label) {
	    $number=($i/1000000);
	    $shift=length($number)*$w/2;
	    $position=(($i-$beg)*($x-2*$border)/$seq_len)+$border-$shift;
	    $graph->string($font,$position,0,$number,$black);
	};
# add MB at and
#  $position += 2*$w;
	$position = $x-$border;
	$graph->string($font,$position,0,"[Mb]",$black);
    };
};

###############################################################################
# y axis - only labels
sub draw_y_axis {
    my ($y1);
    
    $y1 = $y-(($y_gc_proc_max-$y_gc_proc_min)*$py);
    $graph->string($font,0,$y1,$y_gc_proc_max."%",$black);
    $y1 = $y-(($y_gc_proc_min-$y_gc_proc_min)*$py)-15;
    $graph->string($font,0,$y1,$y_gc_proc_min."%",$black);
};

###############################################################################
# draw graph with colors
sub draw_color_graph {
    my ($points,$points2,$i,$px,$x1,$y1,$color);
    
# points - real; points2 - expected number of points (end of the sequence)
    $points = @data;
    $points2 = ($seq_len/$window-1)*$step;
    $px = ($x-$border*2)/$points2;
    $y_base = $y;
    
    for ($i=1; $i<$points; $i++) {
	$x1 = $border+$px*$i;
	$y1 = $y-(($data[$i]-$y_gc_proc_min)*$py);
	($color) = &resolve_color($data[$i]);
	$graph->line($x1,$y1,$x1,$y_base,$color);
    };
};

###############################################################################
sub resolve_color {
    my $proc = shift;
    my ($i,$color);
# resolve color
    $color = $isochore_color[$isochores];
    for ($i=$isochores-1; $i>=0; $i--) {
	if    ($proc < $isochore_proc[$i]) { 
	    $color = $isochore_color[$i];
	};
    };
    return $color;
};

###############################################################################
# draw graph
sub draw_graph {
    my ($points,$points2,$i,$px,$x1,$x2,$y1,$y2);
    
# points - real; points2 - expected number of points (end of the sequence)
    $points = @data;
    $points2 = ($seq_len/$window-1)*$step;
    $px = ($x-$border*2)/$points2;
    
    $x1 = $border;
    $y1=$y-(($data[0]-$y_gc_proc_min)*$py);
    for ($i=1; $i<$points; $i++) {
	$x2 = $x1;
	$y2 = $y1;
	$x1 = $border+$px*$i;
	$y1 = $y-(($data[$i]-$y_gc_proc_min)*$py);
	$graph->line($x1,$y1,$x2,$y2,$black);
    };
};

###############################################################################
# shade unknown sequence
sub shade_gaps {
    my ($px,$y1,$y2,$x1,$x2,$i);
    
    $px = ($x-$border*2)/$seq_len;
    $y1 = $y-($y_gc_proc_max-$y_gc_proc_min)*$py;
    $y2 = $y;

    for ($i=0; $i<$gap_number; $i++) {
	$x1 = $border+$gap_beg[$i]*$px;
	$x2 = $border+$gap_end[$i]*$px;
	$graph->filledRectangle($x1,$y1,$x2,$y2,$grey);
    };
};

###############################################################################
# draw isochores
sub draw_isochores {
    my ($i,$y1);
    
# boxes
    $graph->filledRectangle($x-$border/5*4,$y_gc_proc_max-$y_gc_proc_min,$x-$border/5,$y,$isochore_color[$isochores]);
    for ($i=$isochores-1; $i>=0; $i--) {
	$y1 = $y-(($isochore_proc[$i]-$y_gc_proc_min)*$py);
	$graph->filledRectangle($x-$border/5*4,$y1,$x-$border/5,$y,$isochore_color[$i]);
    };
    
# description
    if ($labels) {
	$y1 = $y-(($y_gc_proc_min+$isochore_proc[0])/2-$y_gc_proc_min)*$py;
	$graph->string($font,$x-$border/2-$w,$y1-$h/2,$isochore_name[0],$black);
	for ($i=1; $i<$isochores; $i++) {
	    $y1 = $y-(($isochore_proc[$i-1]+$isochore_proc[$i])/2-$y_gc_proc_min)*$py;
	    $graph->string($font,$x-$border/2-$w,$y1-$h/2,$isochore_name[$i],$black);
	};
	$y1 = $y-(($isochore_proc[$isochores-1]+$y_gc_proc_max)/2-$y_gc_proc_min)*$py;
	$graph->string($font,$x-$border/2-$w,$y1-$h/2,$isochore_name[$isochores],$black);
    };

# lines  
    $graph->setStyle($black,gdTransparent);
    for ($i=0; $i<$isochores; $i++) {
	$y1 = $y-(($isochore_proc[$i]-$y_gc_proc_min)*$py);
	$graph->string($font,0,$y1-$h/2,$isochore_proc[$i]."%",$black) if $labels;
	$graph->dashedLine($border,$y1,$x-$border,$y1,gdStyled);
    };
    
};

###############################################################################
# draw labels
sub draw_ch_name {
    my ($font,$w,$h,$x1,$y1);
    
# chromosome name
    $font = gdMediumBoldFont;
#  $font = gdSmallFont;
#  $font = gdLargeFont;
    $w=$font->width; $h=$font->height;
    $x1 = (length($ch_name)*$w+8)/2; 
    $y1 = ($h+8)/2;
    
    $graph->arc($x1+1,$y1+1,$x1*2,$y1*2,0,360,$black); 
    $graph->fill($x1+1,$y1+1,$black);
    $graph->string($font,$x1-length($ch_name)*$w/2+1,$y1-$h/2+1,$ch_name,$white);
};

###############################################################################
# end
###############################################################################
